import { inject, isSignal, signal, computed, effect } from '@angular/core';
import { resolveRef } from 'angular-three';
import { NgtcPhysics } from 'angular-three-cannon';
import { NgtcDebug } from 'angular-three-cannon/debug';
import { assertInjector } from 'ngxtension/assert-injector';
import { Euler, Quaternion, Vector3, InstancedMesh, DynamicDrawUsage, Object3D } from 'three';

function capitalize(str) {
    return (str.charAt(0).toUpperCase() + str.slice(1));
}
function getUUID(body, index) {
    const suffix = index === undefined ? '' : `/${index}`;
    if (typeof body === 'function')
        return null;
    return body && body && `${body.uuid}${suffix}`;
}
const e = new Euler();
const q = new Quaternion();
function quaternionToRotation(callback) {
    return (v) => callback(e.setFromQuaternion(q.fromArray(v)).toArray());
}
let incrementingId = 0;
function createSubscribe(body, worker, subscriptions, type, index, target = 'bodies') {
    return (callback) => {
        const id = incrementingId++;
        subscriptions[id] = { [type]: callback };
        const uuid = getUUID(body, index);
        uuid && worker.subscribe({ props: { id, target, type }, uuid });
        return () => {
            delete subscriptions[id];
            worker.unsubscribe({ props: id });
        };
    };
}
function makeTriplet(v) {
    return v instanceof Vector3 ? [v.x, v.y, v.z] : v;
}
function prepare(object, { position = [0, 0, 0], rotation = [0, 0, 0], userData = {} }) {
    object.userData = userData;
    object.position.set(...position);
    object.rotation.set(...rotation);
    object.updateMatrix();
}
function setupCollision(events, { onCollide, onCollideBegin, onCollideEnd }, uuid) {
    events[uuid] = { collide: onCollide, collideBegin: onCollideBegin, collideEnd: onCollideEnd };
}
function makeBodyApi(body, worker, { subscriptions, scaleOverrides }) {
    const makeAtomic = (type, index) => {
        const op = `set${capitalize(type)}`;
        return {
            set: (value) => {
                const uuid = getUUID(body, index);
                uuid && worker[op]({ props: value, uuid });
            },
            subscribe: createSubscribe(body, worker, subscriptions, type, index),
        };
    };
    const makeQuaternion = (index) => {
        const type = 'quaternion';
        return {
            copy: ({ w, x, y, z }) => {
                const uuid = getUUID(body, index);
                uuid && worker.setQuaternion({ props: [x, y, z, w], uuid });
            },
            set: (x, y, z, w) => {
                const uuid = getUUID(body, index);
                uuid && worker.setQuaternion({ props: [x, y, z, w], uuid });
            },
            subscribe: createSubscribe(body, worker, subscriptions, type, index),
        };
    };
    const makeRotation = (index) => {
        return {
            copy: ({ x, y, z }) => {
                const uuid = getUUID(body, index);
                uuid && worker.setRotation({ props: [x, y, z], uuid });
            },
            set: (x, y, z) => {
                const uuid = getUUID(body, index);
                uuid && worker.setRotation({ props: [x, y, z], uuid });
            },
            subscribe: (callback) => {
                const id = incrementingId++;
                const target = 'bodies';
                const type = 'quaternion';
                const uuid = getUUID(body, index);
                subscriptions[id] = { [type]: quaternionToRotation(callback) };
                uuid && worker.subscribe({ props: { id, target, type }, uuid });
                return () => {
                    delete subscriptions[id];
                    worker.unsubscribe({ props: id });
                };
            },
        };
    };
    const makeVec = (type, index) => {
        const op = `set${capitalize(type)}`;
        return {
            copy: ({ x, y, z }) => {
                const uuid = getUUID(body, index);
                uuid && worker[op]({ props: [x, y, z], uuid });
            },
            set: (x, y, z) => {
                const uuid = getUUID(body, index);
                uuid && worker[op]({ props: [x, y, z], uuid });
            },
            subscribe: createSubscribe(body, worker, subscriptions, type, index),
        };
    };
    const makeRemove = (index) => {
        const uuid = getUUID(body, index);
        return () => {
            if (uuid)
                worker.removeBodies({ uuid: [uuid] });
        };
    };
    function makeApi(index) {
        return {
            allowSleep: makeAtomic('allowSleep', index),
            angularDamping: makeAtomic('angularDamping', index),
            angularFactor: makeVec('angularFactor', index),
            angularVelocity: makeVec('angularVelocity', index),
            applyForce(force, worldPoint) {
                const uuid = getUUID(body, index);
                uuid && worker.applyForce({ props: [force, worldPoint], uuid });
            },
            applyImpulse(impulse, worldPoint) {
                const uuid = getUUID(body, index);
                uuid && worker.applyImpulse({ props: [impulse, worldPoint], uuid });
            },
            applyLocalForce(force, localPoint) {
                const uuid = getUUID(body, index);
                uuid && worker.applyLocalForce({ props: [force, localPoint], uuid });
            },
            applyLocalImpulse(impulse, localPoint) {
                const uuid = getUUID(body, index);
                uuid && worker.applyLocalImpulse({ props: [impulse, localPoint], uuid });
            },
            applyTorque(torque) {
                const uuid = getUUID(body, index);
                uuid && worker.applyTorque({ props: [torque], uuid });
            },
            collisionFilterGroup: makeAtomic('collisionFilterGroup', index),
            collisionFilterMask: makeAtomic('collisionFilterMask', index),
            collisionResponse: makeAtomic('collisionResponse', index),
            fixedRotation: makeAtomic('fixedRotation', index),
            isTrigger: makeAtomic('isTrigger', index),
            linearDamping: makeAtomic('linearDamping', index),
            linearFactor: makeVec('linearFactor', index),
            mass: makeAtomic('mass', index),
            material: makeAtomic('material', index),
            position: makeVec('position', index),
            quaternion: makeQuaternion(index),
            rotation: makeRotation(index),
            scaleOverride(scale) {
                const uuid = getUUID(body, index);
                if (uuid)
                    scaleOverrides[uuid] = new Vector3(...scale);
            },
            sleep() {
                const uuid = getUUID(body, index);
                uuid && worker.sleep({ uuid });
            },
            sleepSpeedLimit: makeAtomic('sleepSpeedLimit', index),
            sleepTimeLimit: makeAtomic('sleepTimeLimit', index),
            userData: makeAtomic('userData', index),
            velocity: makeVec('velocity', index),
            remove: makeRemove(index),
            wakeUp() {
                const uuid = getUUID(body, index);
                uuid && worker.wakeUp({ uuid });
            },
        };
    }
    const cache = {};
    return { ...makeApi(), at: (index) => cache[index] || (cache[index] = makeApi(index)) };
}
const defaultTransformArgs = {
    Plane: (_) => [],
    Box: (args = [1, 1, 1]) => args,
    Trimesh: (args) => args,
    Cylinder: (_ = []) => [],
    Heightfield: (args) => args,
    ConvexPolyhedron: ([vertices, faces, normals, axes, boundingSphereRadius] = []) => [
        vertices && vertices.map(makeTriplet),
        faces,
        normals && normals.map(makeTriplet),
        axes && axes.map(makeTriplet),
        boundingSphereRadius,
    ],
    Particle: (_) => [],
    Sphere: (args = [1]) => {
        if (!Array.isArray(args))
            throw new Error('Sphere body args must be an array');
        return [args[0]];
    },
    Compound: (args) => args,
};

function createInjectBody(type) {
    return (getPropFn, ref, options) => injectBody(type, getPropFn, ref, options);
}
function injectBody(type, getPropFn, ref, { transformArgs, injector } = {}) {
    return assertInjector(injectBody, injector, () => {
        const physics = inject(NgtcPhysics, { optional: true });
        if (!physics) {
            throw new Error(`[NGT Cannon] injectBody was called outside of <ngtc-physics>`);
        }
        const debug = inject(NgtcDebug, { optional: true });
        const transform = transformArgs ?? defaultTransformArgs[type];
        const isRefSignal = isSignal(ref);
        const bodyRef = (isRefSignal ? ref : signal(ref));
        const body = computed(() => resolveRef(bodyRef()));
        const api = computed(() => {
            const _body = body();
            if (!_body)
                return null;
            const { worker, ...rest } = physics;
            const _worker = worker();
            if (!_worker)
                return null;
            return makeBodyApi(_body, _worker, rest);
        });
        effect((onCleanup) => {
            const currentWorker = physics.worker();
            if (!currentWorker)
                return;
            const object = body();
            if (!isRefSignal && !object) {
                bodyRef.set(resolveRef(ref));
                return;
            }
            if (!object)
                return;
            const [uuid, props] = (() => {
                let uuids = [];
                let temp;
                if (object instanceof InstancedMesh) {
                    object.instanceMatrix.setUsage(DynamicDrawUsage);
                    uuids = new Array(object.count).fill(0).map((_, i) => `${object.uuid}/${i}`);
                    temp = new Object3D();
                }
                else {
                    uuids = [object.uuid];
                }
                return [
                    uuids,
                    uuids.map((id, index) => {
                        const props = getPropFn(index);
                        if (temp) {
                            prepare(temp, props);
                            object.setMatrixAt(index, temp.matrix);
                            object.instanceMatrix.needsUpdate = true;
                        }
                        else {
                            prepare(object, props);
                        }
                        physics.refs[id] = object;
                        debug?.add(id, props, type);
                        setupCollision(physics.events, props, id);
                        // @ts-expect-error - if args is undefined, there's default
                        return { ...props, args: transform(props.args) };
                    }),
                ];
            })();
            // Register on mount, unregister on unmount
            currentWorker.addBodies({
                props: props.map(({ onCollide, onCollideBegin, onCollideEnd, ...serializableProps }) => {
                    return {
                        onCollide: Boolean(onCollide),
                        onCollideBegin: Boolean(onCollideBegin),
                        onCollideEnd: Boolean(onCollideEnd),
                        ...serializableProps,
                    };
                }),
                type,
                uuid,
            });
            onCleanup(() => {
                uuid.forEach((id) => {
                    delete physics.refs[id];
                    debug?.remove(id);
                    delete physics.events[id];
                });
                currentWorker.removeBodies({ uuid });
            });
        });
        return api;
    });
}
const injectBox = createInjectBody('Box');
const injectConvexPolyhedron = createInjectBody('ConvexPolyhedron');
const injectCylinder = createInjectBody('Cylinder');
const injectHeightfield = createInjectBody('Heightfield');
const injectParticle = createInjectBody('Particle');
const injectPlane = createInjectBody('Plane');
const injectSphere = createInjectBody('Sphere');
const injectTrimesh = createInjectBody('Trimesh');
const injectCompound = createInjectBody('Compound');

/**
 * Generated bundle index. Do not edit.
 */

export { injectBox, injectCompound, injectConvexPolyhedron, injectCylinder, injectHeightfield, injectParticle, injectPlane, injectSphere, injectTrimesh };
//# sourceMappingURL=angular-three-cannon-body.mjs.map
