import { inject, isSignal, signal, computed, effect, untracked } from '@angular/core';
import { makeId, resolveRef } from 'angular-three';
import { NgtcPhysics } from 'angular-three-cannon';
import { assertInjector } from 'ngxtension/assert-injector';

function createInjectConstraint(type) {
    return (bodyA, bodyB, options) => injectConstraint(type, bodyA, bodyB, options);
}
function injectConstraint(type, bodyA, bodyB, { injector, options = {}, disableOnStart = false, } = {}) {
    return assertInjector(injectConstraint, injector, () => {
        const physics = inject(NgtcPhysics, { optional: true });
        if (!physics) {
            throw new Error(`[NGT Cannon] injectConstraint was called outside of <ngtc-physics>`);
        }
        const worker = physics.worker;
        const uuid = makeId();
        const bodyARef = isSignal(bodyA) ? bodyA : signal(bodyA);
        const bodyBRef = isSignal(bodyB) ? bodyB : signal(bodyB);
        const bodyAValue = computed(() => resolveRef(bodyARef()));
        const bodyBValue = computed(() => resolveRef(bodyBRef()));
        const constraintApi = computed(() => {
            const _worker = worker();
            if (!_worker)
                return null;
            const enableDisable = {
                disable: () => _worker.disableConstraint({ uuid }),
                enable: () => _worker.enableConstraint({ uuid }),
                remove: () => _worker.removeConstraint({ uuid }),
            };
            if (type === 'Hinge') {
                return {
                    ...enableDisable,
                    disableMotor: () => _worker.disableConstraintMotor({ uuid }),
                    enableMotor: () => _worker.enableConstraintMotor({ uuid }),
                    setMotorMaxForce: (value) => _worker.setConstraintMotorMaxForce({ props: value, uuid }),
                    setMotorSpeed: (value) => _worker.setConstraintMotorSpeed({ props: value, uuid }),
                };
            }
            return enableDisable;
        });
        let alreadyDisabled = false;
        effect((onCleanup) => {
            const currentWorker = worker();
            if (!currentWorker)
                return;
            const [a, b, api] = [bodyAValue(), bodyBValue(), untracked(constraintApi)];
            if (!a || !b)
                return;
            currentWorker.addConstraint({
                props: [a.uuid, b.uuid, options],
                type,
                uuid,
            });
            if (disableOnStart && !alreadyDisabled) {
                alreadyDisabled = true;
                api?.disable();
            }
            onCleanup(() => currentWorker.removeConstraint({ uuid }));
        });
        return constraintApi;
    });
}
const injectPointToPoint = createInjectConstraint('PointToPoint');
const injectConeTwist = createInjectConstraint('ConeTwist');
const injectDistance = createInjectConstraint('Distance');
const injectLock = createInjectConstraint('Lock');
const injectHinge = createInjectConstraint('Hinge');

/**
 * Generated bundle index. Do not edit.
 */

export { injectConeTwist, injectDistance, injectHinge, injectLock, injectPointToPoint };
//# sourceMappingURL=angular-three-cannon-constraint.mjs.map
