import * as i0 from '@angular/core';
import { input, signal, inject, Injector, afterNextRender, effect, untracked, Directive } from '@angular/core';
import { CannonWorkerAPI } from '@pmndrs/cannon-worker-api';
import { injectStore, pick, injectBeforeRender } from 'angular-three';
import { mergeInputs } from 'ngxtension/inject-inputs';
import { Vector3, Quaternion, Matrix4, InstancedMesh } from 'three';

const v = new Vector3();
const s = new Vector3(1, 1, 1);
const q = new Quaternion();
const m = new Matrix4();
function apply(index, positions, quaternions, scale = s, object) {
    if (index !== undefined) {
        m.compose(v.fromArray(positions, index * 3), q.fromArray(quaternions, index * 4), scale);
        if (object) {
            object.matrixAutoUpdate = false;
            object.matrix.copy(m);
        }
        return m;
    }
    return m.identity();
}
function unique() {
    const values = [];
    return (value) => (values.includes(value) ? false : !!values.push(value));
}
const defaultOptions = {
    allowSleep: false,
    axisIndex: 0,
    broadphase: 'Naive',
    defaultContactMaterial: { contactEquationStiffness: 1e6 },
    frictionGravity: null,
    gravity: [0, -9.81, 0],
    isPaused: false,
    iterations: 5,
    maxSubSteps: 10,
    quatNormalizeFast: false,
    quatNormalizeSkip: 0,
    shouldInvalidate: true,
    size: 1000,
    solver: 'GS',
    stepSize: 1 / 60,
    tolerance: 0.001,
};
class NgtcPhysics {
    constructor() {
        this.store = injectStore();
        this.options = input(defaultOptions, { transform: mergeInputs(defaultOptions) });
        this.axisIndex = pick(this.options, 'axisIndex');
        this.broadphase = pick(this.options, 'broadphase');
        this.gravity = pick(this.options, 'gravity');
        this.iterations = pick(this.options, 'iterations');
        this.tolerance = pick(this.options, 'tolerance');
        this.invalidate = this.store.select('invalidate');
        // @ts-expect-error - worker is not nullable, and we don't want to use ! operator.
        this.cannonWorker = signal(null);
        this.bodies = {};
        this.events = {};
        this.refs = {};
        this.scaleOverrides = {};
        this.subscriptions = {};
        this.worker = this.cannonWorker.asReadonly();
        const injector = inject(Injector);
        // NOTE: set new cannonworker in afterNextRender
        // - so inputs are resolved
        // - so the worker is instantiated only once
        // - effects are started after worker is instantiated
        afterNextRender(() => {
            this.cannonWorker.set(new CannonWorkerAPI(this.options()));
            effect((onCleanup) => {
                const cleanup = this.connectWorkerEffect();
                onCleanup(() => cleanup?.());
            }, { injector });
            effect(() => {
                this.updateWorkerStateEffect('axisIndex', this.axisIndex);
                this.updateWorkerStateEffect('broadphase', this.broadphase);
                this.updateWorkerStateEffect('gravity', this.gravity);
                this.updateWorkerStateEffect('iterations', this.iterations);
                this.updateWorkerStateEffect('tolerance', this.tolerance);
            }, { injector });
        });
        let timeSinceLastCalled = 0;
        injectBeforeRender(({ delta }) => {
            const [{ isPaused, maxSubSteps, stepSize }, worker] = [this.options(), this.cannonWorker()];
            if (isPaused || !worker || stepSize == null)
                return;
            timeSinceLastCalled += delta;
            worker.step({ maxSubSteps, stepSize, timeSinceLastCalled });
            timeSinceLastCalled = 0;
        });
    }
    connectWorkerEffect() {
        const worker = this.cannonWorker();
        if (!worker)
            return;
        worker.connect();
        worker.init();
        worker.on('collide', this.collideHandler.bind(this));
        worker.on('collideBegin', this.collideBeginHandler.bind(this));
        worker.on('collideEnd', this.collideEndHandler.bind(this));
        worker.on('frame', this.frameHandler.bind(this));
        worker.on('rayhit', this.rayhitHandler.bind(this));
        return () => {
            worker.terminate();
            worker.removeAllListeners();
        };
    }
    updateWorkerStateEffect(key, option) {
        const worker = this.cannonWorker();
        if (!worker)
            return;
        Object.assign(worker, { [key]: option() });
    }
    collideHandler({ body, contact: { bi, bj, ...contactRest }, target, ...rest }) {
        const { events, refs } = this;
        const cb = events[target]?.collide;
        if (cb) {
            cb({ body: refs[body], contact: { bi: refs[bi], bj: refs[bj], ...contactRest }, target: refs[target], ...rest });
        }
    }
    collideBeginHandler({ bodyA, bodyB }) {
        const { events, refs } = this;
        const cbA = events[bodyA]?.collideBegin;
        if (cbA)
            cbA({ body: refs[bodyB], op: 'event', target: refs[bodyA], type: 'collideBegin' });
        const cbB = events[bodyB]?.collideBegin;
        if (cbB)
            cbB({ body: refs[bodyA], op: 'event', target: refs[bodyB], type: 'collideBegin' });
    }
    collideEndHandler({ bodyA, bodyB }) {
        const { events, refs } = this;
        const cbA = events[bodyA]?.collideEnd;
        if (cbA)
            cbA({ body: refs[bodyB], op: 'event', target: refs[bodyA], type: 'collideEnd' });
        const cbB = events[bodyB]?.collideEnd;
        if (cbB)
            cbB({ body: refs[bodyA], op: 'event', target: refs[bodyB], type: 'collideEnd' });
    }
    frameHandler({ active, bodies: uuids = [], observations, positions, quaternions, }) {
        const [{ shouldInvalidate }, { bodies, subscriptions, refs, scaleOverrides }, invalidate] = [
            untracked(this.options),
            this,
            this.invalidate(),
        ];
        for (let i = 0; i < uuids.length; i++) {
            bodies[uuids[i]] = i;
        }
        observations.forEach(([id, value, type]) => {
            const subscription = subscriptions[id] || {};
            const cb = subscription[type];
            // @ts-expect-error - We clearly know the type of the callback, but typescript can't deal with it
            cb && cb(value);
        });
        if (!active)
            return;
        for (const ref of Object.values(refs).filter(unique())) {
            if (ref instanceof InstancedMesh) {
                for (let i = 0; i < ref.count; i++) {
                    const uuid = `${ref.uuid}/${i}`;
                    const index = bodies[uuid];
                    if (index !== undefined) {
                        ref.setMatrixAt(i, apply(index, positions, quaternions, scaleOverrides[uuid]));
                        ref.instanceMatrix.needsUpdate = true;
                    }
                }
            }
            else {
                const scale = scaleOverrides[ref.uuid] || ref.scale;
                apply(bodies[ref.uuid], positions, quaternions, scale, ref);
            }
        }
        if (shouldInvalidate)
            invalidate();
    }
    rayhitHandler({ body, ray: { uuid, ...rayRest }, ...rest }) {
        const { events, refs } = this;
        const cb = events[uuid]?.rayhit;
        if (cb)
            cb({ body: body ? refs[body] : null, ray: { uuid, ...rayRest }, ...rest });
    }
    static { this.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "19.1.4", ngImport: i0, type: NgtcPhysics, deps: [], target: i0.ɵɵFactoryTarget.Directive }); }
    static { this.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "17.1.0", version: "19.1.4", type: NgtcPhysics, isStandalone: true, selector: "ngtc-physics", inputs: { options: { classPropertyName: "options", publicName: "options", isSignal: true, isRequired: false, transformFunction: null } }, ngImport: i0 }); }
}
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "19.1.4", ngImport: i0, type: NgtcPhysics, decorators: [{
            type: Directive,
            args: [{ selector: 'ngtc-physics' }]
        }], ctorParameters: () => [] });

/**
 * Generated bundle index. Do not edit.
 */

export { NgtcPhysics };
//# sourceMappingURL=angular-three-cannon.mjs.map
